const _null = null, _undefined = undefined, _Infinity = Infinity;
const _window = window, { document: _document } = _window, _Object = Object, { assign: _assign } = _Object, _Map = Map, _Set = Set, _Math = Math, { sqrt: _sqrt, pow: _pow, sin: _sin, cos: _cos, PI: _PI, max: _max, min: _min, abs: _abs, sign: _sign } = _Math, { now: _now } = Date, _setTimeout = setTimeout, _clearTimeout = clearTimeout;

const Utils = {
    Const: {
        EMPTY_OBJECT: _Object.create(_null),
        DOUBLE_PI: _PI * 2,
        HALF_PI: _PI / 2,
        TRANSPARENT: 'rgba(0,0,0,0)',
        IS_TOUCH_MODE: navigator.maxTouchPoints > 0 || /iOS|iPhone|iPod|iPad/.test(navigator.userAgent),
    },
    removeIndex(array, index) {
        if (!index) {
            array.shift();
        }
        else {
            const end = array.length - 1;
            if (index < end) {
                for (; index < end; index++) {
                    array[index] = array[index + 1];
                }
                array.length--;
            }
            else {
                array.pop();
            }
        }
    },
    insert(array, index, item) {
        if (!index) {
            array.unshift(item);
        }
        else if (index < array.length) {
            const end = array.length++ - 1;
            for (let i = end; i >= index; i--) {
                array[i + 1] = array[i];
            }
            array[index] = item;
        }
        else {
            array.push(item);
        }
    },
    mix(a, b, k) {
        if (a === _Infinity) {
            if (b === a) {
                return a;
            }
            else if (b === -a) {
                return NaN;
            }
        }
        else if (a === -_Infinity) {
            if (b === a) {
                return a;
            }
            else if (b === -a) {
                return NaN;
            }
        }
        return a + (b - a) * k;
    },
    random(min, max) {
        return Utils.mix(min, max, _Math.random());
    },
    pick(options) {
        return options[_Math.floor(options.length * _Math.random())];
    },
    rad2deg(rad) {
        return rad / _PI * 180;
    },
    deg2rad(deg) {
        return deg / 180 * _PI;
    },
    quadraticSum(a, b) {
        return a * a + b * b;
    },
    distance(x1, y1, x2, y2) {
        return _sqrt(Utils.quadraticSum(x2 - x1, y2 - y1));
    },
    debounce: _assign(function debounce(callback, delay) {
        let timer;
        const wrapper = function debounceWrapper() {
            if (timer !== _undefined) {
                _clearTimeout(timer);
            }
            timer = _setTimeout.apply(_window, [callback, wrapper.delay].concat(arguments));
        };
        wrapper.delay = delay || Utils.debounce.defaultDelay;
        return wrapper;
    }, {
        defaultDelay: 100
    }),
    throttle: _assign(function throttle(callback, threshold) {
        const wrapper = function throttleWrapper() {
            const currentTime = _now();
            if (currentTime - wrapper.lastCallTime >= wrapper.threshold) {
                wrapper.lastCallTime = currentTime;
                return callback.apply(this, arguments);
            }
        };
        wrapper.lastCallTime = 0;
        wrapper.threshold = threshold || Utils.throttle.defaultThreshold;
        return wrapper;
    }, {
        defaultThreshold: 100
    }),
    cache(fn) {
        const map = new _Map();
        const wrapper = function cacheWrapper(...args) {
            const parameters = args.join();
            if (map.has(parameters)) {
                return map.get(parameters);
            }
            else {
                const result = fn.apply(this, args);
                map.set(parameters, result);
                return result;
            }
        };
        wrapper.cache = map;
        return wrapper;
    },
};

class EventEmitter {
    constructor() {
        this._listenerMap = new _Map();
        this.on = this.on.bind(this);
        this.off = this.off.bind(this);
        this.once = this.once.bind(this);
        this.emit = this.emit.bind(this);
        this.clearEvent = this.clearEvent.bind(this);
        this.clearEvents = this.clearEvents.bind(this);
    }
    on(event, listener, once) {
        const { _listenerMap } = this;
        if (_listenerMap.has(event)) {
            _listenerMap.get(event).push([listener, !once]);
        }
        else {
            _listenerMap.set(event, [[listener, !once]]);
        }
        return this;
    }
    off(event, listener, once) {
        once = !once;
        const { _listenerMap } = this;
        if (_listenerMap.has(event)) {
            const records = _listenerMap.get(event), index = records.findIndex(record => record[0] === listener &&
                record[1] === once);
            if (!~index) {
                Utils.removeIndex(records, index);
            }
        }
        return this;
    }
    once(event, listener) {
        return this.on(event, listener, true);
    }
    emit(event, ...args) {
        const { _listenerMap } = this;
        if (_listenerMap.has(event)) {
            const records = _listenerMap.get(event);
            if (records.length) {
                _listenerMap.set(event, records.filter(record => {
                    record[0].apply(this, args);
                    return record[1];
                }));
                return true;
            }
        }
        return false;
    }
    clearEvent(event) {
        this._listenerMap.delete(event);
        return this;
    }
    clearEvents() {
        this._listenerMap.clear();
        return this;
    }
}

class Pool {
    constructor(factory, options) {
        this.factory = factory;
        this.options = options;
        this._items = new Array();
    }
    get size() {
        return this._items.length;
    }
    set size(value) {
        const { _items } = this, { length: currentSize } = _items;
        _items.length = value;
        if (currentSize < value) {
            for (let i = currentSize; i < value; i++) {
                _items[i] = new this.factory(this.options);
            }
        }
    }
    get() {
        const { _items } = this;
        if (_items.length) {
            return _items.shift();
        }
        else {
            return new this.factory(this.options);
        }
    }
    add(item) {
        this._items.push(item);
        return this;
    }
}

const interpolate = (p1, p2, t) => {
    const _t = 1 - t;
    return 3 * p1 * _t * _t * t + 3 * p2 * t * t * _t + t * t * t;
};
const cubic = (x1, y1, x2, y2) => x => {
    if (x <= 0) {
        return 0;
    }
    else if (x >= 1) {
        return 1;
    }
    const { accuracy } = cubic;
    let l = 0, r = 1, t = .5, cur;
    while (_abs((cur = (interpolate(x1, x2, t))) - x) > accuracy) {
        if (cur > x) {
            r = t;
        }
        else {
            l = t;
        }
        t = (l + r) / 2;
    }
    return interpolate(y1, y2, t);
};
cubic.accuracy = .001;
const Timing = {
    linear: ((x) => x),
    cubic,
    ease: cubic(.25, .2, .25, 1),
    easeIn: cubic(.42, 0, 1, 1),
    easeOut: cubic(0, 0, .58, 1),
    easeInOut: cubic(.42, 0, 0.25, 1),
    steps(stepCount, start) {
        const step = 1 / stepCount;
        return x => {
            let t = _Math.floor(x / step);
            if (start && t < stepCount) {
                t += 1;
            }
            return t * step;
        };
    },
};

class Transition extends EventEmitter {
    constructor(options = Utils.Const.EMPTY_OBJECT) {
        super();
        this._delay = 0;
        _assign(this, Transition.defaults, options);
        if (this.active) {
            this._startTime = _now();
        }
        const { target, key } = this;
        if (options.from === _undefined && target) {
            this.from = target[key];
        }
        if (options.to === _undefined && target) {
            this.to = target[key];
        }
    }
    start() {
        this._startTime = _now();
        this.active = true;
        this._delay = 0;
        this.emit('start');
    }
    update() {
        if (!this.active) {
            return;
        }
        const { from, to } = this;
        let x = (_now() - this._startTime - this.delay - this._delay) / this.duration, value;
        if (x >= 1) {
            value = to;
        }
        else if (x <= 0) {
            value = from;
        }
        else {
            value = Utils.mix(from, to, this.timing(x));
        }
        const { target } = this;
        if (target) {
            target[this.key] = value;
        }
        this.emit('update', value);
        if (x >= 1) {
            this.active = false;
            this.emit('end');
            if (this.loop) {
                this.start();
            }
        }
    }
    pause() {
        this.active = false;
        this._pauseTime = _now();
    }
    resume() {
        this._delay += _now() - this._pauseTime;
        this.active = true;
    }
    finish() {
        this.emit('update', this.to);
        this.active = false;
        this.emit('end');
    }
}
Transition.defaults = {
    active: true,
    from: 0,
    to: 1,
    duration: 1000,
    delay: 0,
    loop: false,
    timing: Timing.linear,
    target: _null,
};

// TODO: add interface `ReadonlyVector`
class Vector {
    static of(x, y) {
        return new Vector(x, y);
    }
    static from(vectorLike) {
        return Vector.of(vectorLike.x, vectorLike.y);
    }
    static plus(vectors) {
        return vectors.reduce((ans, cur) => ans.plusVector(cur), new Vector());
    }
    static mean(vectors) {
        return Vector.plus(vectors).shrink(vectors.length);
    }
    static dot(vector1, vector2) {
        return vector1.x * vector2.x + vector1.y * vector2.y;
    }
    static cross(vector1, vector2) {
        return vector1.x * vector2.y - vector1.y * vector2.x;
    }
    static minus(vector1, vector2) {
        return Vector.of(vector1.x - vector2.x, vector1.y - vector2.y);
    }
    static project(sourceVector, directionVector) {
        const dirVecMod = directionVector.getNorm();
        return dirVecMod && Vector.dot(sourceVector, directionVector) / dirVecMod;
    }
    static projectVector(sourceVector, directionVector) {
        return directionVector.clone().setNorm(Vector.project(sourceVector, directionVector));
    }
    static distance(vector1, vector2) {
        return Utils.distance(vector1.x, vector1.y, vector2.x, vector2.y);
    }
    static distribute(vector0, vector1, vector2, k1, k2, scale) {
        let sum = _abs(k1) + _abs(k2);
        if (scale !== _undefined) {
            sum /= scale;
        }
        k1 /= sum;
        k2 /= sum;
        vector1.x += vector0.x * k1;
        vector1.y += vector0.y * k1;
        vector2.x += vector0.x * k2;
        vector2.y += vector0.y * k2;
    }
    static random(startAngle, endAngle = Utils.Const.DOUBLE_PI) {
        const angle = Utils.random(startAngle || 0, endAngle);
        return Vector.of(_cos(angle), _sin(angle));
    }
    constructor(x, y) {
        this.x = x || 0;
        this.y = y || 0;
    }
    clone() {
        return Vector.of(this.x, this.y);
    }
    set(x, y) {
        this.x = x;
        this.y = y;
        return this;
    }
    setVector(vector, scale) {
        if (scale !== _undefined) {
            this.x = vector.x * scale;
            this.y = vector.y * scale;
        }
        else {
            this.x = vector.x;
            this.y = vector.y;
        }
        return this;
    }
    reset() {
        this.x = this.y = 0;
        return this;
    }
    getNorm() {
        return _sqrt(_pow(this.x, 2) + _pow(this.y, 2));
    }
    setNorm(norm) {
        const currentNorm = this.getNorm();
        if (currentNorm) {
            const scale = norm / currentNorm;
            this.x *= scale;
            this.y *= scale;
        }
        return this;
    }
    plus(dx, dy) {
        this.x += dx;
        this.y += dy;
        return this;
    }
    plusVector(vector, scale) {
        if (scale !== _undefined) {
            this.x += vector.x * scale;
            this.y += vector.y * scale;
        }
        else {
            this.x += vector.x;
            this.y += vector.y;
        }
        return this;
    }
    minus(dx, dy) {
        this.x -= dx;
        this.y -= dy;
        return this;
    }
    minusVector(vector, scale) {
        if (scale !== _undefined) {
            this.x -= vector.x * scale;
            this.y -= vector.y * scale;
        }
        else {
            this.x -= vector.x;
            this.y -= vector.y;
        }
        return this;
    }
    scale(scaleX, scaleY = scaleX, origin) {
        if (origin) {
            this.x = origin.x + (this.x - origin.x) * scaleX;
            this.y = origin.y + (this.y - origin.y) * scaleY;
        }
        else {
            this.x *= scaleX;
            this.y *= scaleY;
        }
        return this;
    }
    shrink(shrinkX, shrinkY = shrinkX, origin) {
        if (origin) {
            this.x = origin.x + (this.x - origin.x) / shrinkX;
            this.y = origin.y + (this.y - origin.y) / shrinkY;
        }
        else {
            this.x /= shrinkX;
            this.y /= shrinkY;
        }
        return this;
    }
    grow(delta) {
        const norm = this.getNorm();
        if (norm) {
            const scale = (norm + delta) / norm;
            this.x *= scale;
            this.y *= scale;
        }
        return this;
    }
    rotate(radian, origin) {
        const cos = _cos(radian), sin = _sin(radian);
        if (origin) {
            const dx = this.x - origin.x, dy = this.y - origin.y;
            this.x = origin.x + dx * cos - dy * sin;
            this.y = origin.y + dx * sin + dy * cos;
        }
        else {
            const { x, y } = this;
            this.x = x * cos - y * sin;
            this.y = x * sin + y * cos;
        }
        return this;
    }
    getAngle() {
        const { x } = this;
        return _Math.atan(this.y / x) + (x > 0 ? 0 : _PI);
    }
    turn(clockwise) {
        const { x, y } = this;
        if (clockwise) {
            this.x = y;
            this.y = -x;
        }
        else {
            this.x = -y;
            this.y = x;
        }
        return this;
    }
    reverse() {
        this.x = -this.x;
        this.y = -this.y;
        return this;
    }
    normalize() {
        const norm = this.getNorm();
        this.x /= norm;
        this.y /= norm;
        return this;
    }
    isZero() {
        return this.x === 0 && this.y === 0;
    }
    toString(fractionDigits) {
        return this.x.toFixed(fractionDigits) + ',' + this.y.toFixed(fractionDigits);
    }
}

class Bounds {
    constructor() {
        this.top = 0;
        this.right = 0;
        this.bottom = 0;
        this.left = 0;
    }
    get width() {
        return this.right - this.left;
    }
    get height() {
        return this.bottom - this.top;
    }
    move(deltaX, deltaY) {
        this.left += deltaX;
        this.right += deltaX;
        this.top += deltaY;
        this.bottom += deltaY;
    }
    moveVector(vector, scale) {
        if (scale) {
            this.left += vector.x * scale;
            this.right += vector.x * scale;
            this.top += vector.y * scale;
            this.bottom += vector.y * scale;
        }
        else {
            this.left += vector.x;
            this.right += vector.x;
            this.top += vector.y;
            this.bottom += vector.y;
        }
    }
    grow(delta) {
        this.left -= delta;
        this.right += delta;
        this.top -= delta;
        this.bottom += delta;
    }
    overlaps(bounds) {
        return !(bounds.right < this.left ||
            bounds.left > this.right ||
            bounds.bottom < this.top ||
            bounds.top > this.bottom);
    }
    updateVertices(vertices) {
        vertices.forEach(({ x, y }, i) => {
            if (i > 0) {
                if (x < this.left) {
                    this.left = x;
                }
                else if (x > this.right) {
                    this.right = x;
                }
                if (y < this.top) {
                    this.top = y;
                }
                else if (y > this.bottom) {
                    this.bottom = y;
                }
            }
            else {
                this.left = this.right = x;
                this.top = this.bottom = y;
            }
        });
    }
    updateBounds(boundsArray) {
        boundsArray.forEach(({ left, right, top, bottom }, i) => {
            if (i > 0) {
                if (left < this.left) {
                    this.left = left;
                }
                if (right > this.right) {
                    this.right = right;
                }
                if (top < this.top) {
                    this.top = top;
                }
                if (bottom > this.bottom) {
                    this.bottom = bottom;
                }
            }
            else {
                this.left = left;
                this.right = right;
                this.top = top;
                this.bottom = bottom;
            }
        });
    }
}

const Vertices = {
    fromArray(array) {
        const vertices = [];
        for (let i = 0; i < array.length; i += 2) {
            vertices.push(Vector.of(array[i], array[i + 1]));
        }
        return vertices;
    },
    findClosest(target, vertices) {
        let min = _Infinity, closest;
        vertices.forEach(vertex => {
            const current = Utils.quadraticSum(target.x - vertex.x, target.y - vertex.y);
            if (current < min) {
                min = current;
                closest = vertex;
            }
        });
        return closest;
    },
    createPolygon: Utils.cache(function (edges, radius, rotation = -Utils.Const.HALF_PI) {
        const angle = Utils.Const.DOUBLE_PI / edges, results = [];
        for (let i = 0; i < edges; i++) {
            results.push(Vector.of(_cos(rotation), _sin(rotation)).setNorm(radius));
            rotation += angle;
        }
        return results;
    }),
    createStar: Utils.cache(function (angles, innerRadius, outerRadius, rotation = -Utils.Const.HALF_PI) {
        const angle = _PI / angles, results = [];
        for (let i = 0; i < angles; i++) {
            results.push(Vector.of(_cos(rotation), _sin(rotation)).setNorm(innerRadius));
            rotation += angle;
            results.push(Vector.of(_cos(rotation), _sin(rotation)).setNorm(outerRadius));
            rotation += angle;
        }
        return results;
    }),
    createRectangle: Utils.cache(function (width, height, rotation) {
        const x0 = width / 2, y0 = height / 2;
        if (rotation) {
            const cos = _cos(rotation), sin = _sin(rotation), x1 = x0 * cos - y0 * sin, y1 = x0 * sin + y0 * sin, x2 = -x0 * cos - y0 * sin, y2 = -x0 * sin + y0 * sin;
            return Vertices.fromArray([
                x1, y1,
                x2, y2,
                -x1, -y1,
                -x2, -y2
            ]);
        }
        else {
            return Vertices.fromArray([
                x0, y0,
                -x0, y0,
                -x0, -y0,
                x0, -y0
            ]);
        }
    }),
};

const Category = {
    MAX_COUNT: 32,
    FULL_MASK: 4294967295,
    registry: [],
    count: 0,
    next() {
        if (Category.count > Category.MAX_COUNT) {
            throw "Max category count exceeded";
        }
        return 1 << Category.count++;
    },
    for(tag) {
        const record = Category.registry.find(record => record[0] === tag);
        if (!record) {
            const category = Category.next();
            Category.registry.push([tag, category]);
            return category;
        }
        else {
            return record[1];
        }
    },
    tagFor(category) {
        const result = Category.registry.find(record => record[1] === category);
        return result && result[0];
    },
};

class Body extends EventEmitter {
    constructor(options = Utils.Const.EMPTY_OBJECT) {
        super();
        this.tag = '';
        this.bounds = new Bounds();
        this.area = 0;
        this.normals = [];
        this.scaleX = 1;
        this.scaleY = 1;
        this.rotation = 0;
        this.mass = 0;
        this.speed = 0;
        this.contact = new _Set();
        _assign(this, Body.defaults, options);
        if (!this.position) {
            this.position = new Vector();
        }
        if (!this.velocity) {
            this.velocity = new Vector();
        }
        if (!this.acceleration) {
            this.acceleration = new Vector();
        }
        if (this.category) {
            this.tag = Category.tagFor(this.category) || '';
        }
        else if (this.tag) {
            this.category = Category.for(this.tag);
        }
        const { scaleX, scaleY, rotation } = this;
        if (scaleX !== _undefined || scaleY !== _undefined) {
            this.scale(scaleX, scaleY);
        }
        if (rotation) {
            this.rotate(rotation);
        }
        if (this.draggable) {
            this.interactive = true;
        }
        if (this.friction > this.staticFriction) {
            this.friction = this.staticFriction;
        }
    }
    get x() {
        return this.position.x;
    }
    set x(value) {
        const { position } = this;
        this.bounds.move(value - position.x, 0);
        position.x = value;
    }
    get y() {
        return this.position.y;
    }
    set y(value) {
        const { position } = this;
        this.bounds.move(0, value - position.y);
        position.y = value;
    }
    setDensity(density) {
        this.density = density;
        if (this.active) {
            this.mass = this.area * density;
        }
        return this;
    }
    setMass(mass) {
        if (this.active) {
            this.mass = mass;
        }
        this.density = mass / this.area;
        return this;
    }
    _setArea(area) {
        this.area = area;
        this.mass = this.density * area;
    }
    activate(active = true) {
        this.mass = (this.active = active) ? this.area * this.density : _Infinity;
        return this;
    }
    setScale(scaleX, scaleY = scaleX) {
        const deltaScaleX = scaleX / this.scaleX, deltaScaleY = scaleY / this.scaleY, scale = deltaScaleX * deltaScaleY;
        this.area *= scale;
        this.mass *= scale;
        if (this._scale) {
            this._scale(deltaScaleX, deltaScaleY);
        }
        this.scaleX = scaleX;
        this.scaleY = scaleY;
        return this;
    }
    scale(scaleX, scaleY = scaleX, origin) {
        const scale = scaleX * scaleY;
        this.area *= scale;
        this.mass *= scale;
        if (this._scale) {
            this._scale(scaleX, scaleY, origin);
        }
        this.scaleX *= scaleX;
        this.scaleY *= scaleY;
        if (origin) {
            this.position.scale(scaleX, scaleY, origin);
        }
        return this;
    }
    shrink(shrinkX, shrinkY = shrinkX, origin) {
        const shrink = shrinkX * shrinkY;
        this.area /= shrink;
        this.mass /= shrink;
        if (this._scale) {
            this._scale(1 / shrinkX, 1 / shrinkY, origin);
        }
        this.scaleX /= shrinkX;
        this.scaleY /= shrinkY;
        if (origin) {
            this.position.shrink(shrinkX, shrinkY, origin);
        }
        return this;
    }
    setRotation(rotation, origin) {
        const deltaRotation = rotation - this.rotation;
        if (this._rotate) {
            this._rotate(deltaRotation, origin);
        }
        this.rotation = rotation;
        if (origin) {
            this.position.rotate(deltaRotation, origin);
        }
        return this;
    }
    rotate(rotation, origin) {
        if (this._rotate) {
            this._rotate(rotation, origin);
        }
        this.rotation += rotation;
        if (origin) {
            this.position.rotate(rotation, origin);
        }
        return this;
    }
    applyForce(force, point) {
        const { mass } = this;
        this.acceleration.plus(force.x / mass, force.y / mass);
        // TODO: update angular speed
        return this;
    }
    move(deltaX, deltaY) {
        this.position.plus(deltaX, deltaY);
        this.bounds.move(deltaX, deltaY);
        return this;
    }
    moveVector(vector, scale) {
        this.position.plusVector(vector, scale);
        this.bounds.moveVector(vector, scale);
        return this;
    }
    moveTo(x, y) {
        const { position } = this, deltaX = x - position.x, deltaY = y - position.y;
        this.position.plus(deltaX, deltaY);
        this.bounds.move(deltaX, deltaY);
        return this;
    }
    moveToVector(vector) {
        const deltaVector = Vector.minus(vector, this.position);
        this.position.plusVector(deltaVector);
        this.bounds.moveVector(deltaVector);
        return this;
    }
    update(timeScale) {
        this.emit('willUpdate', timeScale);
        if (this.active) {
            const { velocity, bounds, position } = this, maxSpeed = this.maxSpeed;
            let speed = velocity.getNorm() / timeScale;
            if (speed > maxSpeed) {
                velocity.scale(maxSpeed / speed);
                speed = maxSpeed;
            }
            position.plusVector(velocity, timeScale);
            bounds.moveVector(velocity, timeScale);
            this.isStatic = (this.speed = speed) <= Body.maxStaticSpeed;
            const { acceleration, gravity } = this, maxAngularSpeed = this.maxAngularSpeed * timeScale, airSpeedScale = 1 - this.airFriction, angularSpeed = (this.angularSpeed *= airSpeedScale) * timeScale;
            velocity.plusVector(acceleration, timeScale);
            if (gravity) {
                velocity.plusVector(gravity, timeScale);
            }
            velocity.scale(airSpeedScale);
            acceleration.reset();
            if (angularSpeed > maxAngularSpeed) {
                this.rotation += (this.angularSpeed = maxAngularSpeed);
            }
            else {
                this.rotation += angularSpeed;
            }
            if (this.fixRotation) {
                this.rotation %= Utils.Const.DOUBLE_PI;
            }
        }
        else {
            this.isStatic = false;
        }
        this.emit('didUpdate', timeScale);
    }
}
Body.normalPrecision = 3;
Body.maxStaticSpeed = 1;
Body.defaults = {
    category: 0,
    collisionFilter: Category.FULL_MASK,
    sensorFilter: 0,
    isCircle: false,
    active: false,
    interactive: false,
    draggable: false,
    defer: false,
    maxSpeed: 200,
    angularSpeed: 0,
    maxAngularSpeed: 100,
    gravity: Vector.of(0, 2),
    density: 1,
    stiffness: 1,
    slop: .1,
    elasticity: .3,
    friction: .4,
    staticFriction: .5,
    airFriction: 0,
    fixRotation: true,
    radius: 0,
};

const Collision = {
    check(bodies, checker) {
        const { maxStaticSpeed } = Body;
        Collision.find(bodies, checker).filter(collisionInfo => {
            const { body1, body2, overlapVector } = collisionInfo, { velocity: v1, stiffness: stiffness1, slop: slop1, mass: m1, elasticity: elasticity1 } = body1;
            overlapVector.reverse();
            body1.emit('collision', body2, collisionInfo);
            overlapVector.reverse();
            body2.emit('collision', body1, collisionInfo);
            if (body1.category & body2.sensorFilter || body1.sensorFilter & body2.category) {
                return false;
            }
            const { velocity: v2, mass: m2 } = body2, elasticity = _min(elasticity1, body2.elasticity) + 1, slop = slop1 + body2.slop, impulse = collisionInfo.overlap * (stiffness1 + body2.stiffness) / 2, impulseScale = impulse > slop ? (impulse - slop) / impulse : 0, { edgeVector, relativeVelocity } = collisionInfo, isSeparating = Vector.dot(relativeVelocity, overlapVector) < 0, relativeNormalVelocity = Vector.projectVector(relativeVelocity, overlapVector);
            if (body1.active) {
                if (body2.active) {
                    if (impulseScale) {
                        v1.minusVector(overlapVector, impulseScale / 2);
                        v2.plusVector(overlapVector, impulseScale / 2);
                    }
                    if (edgeVector && isSeparating) {
                        Vector.distribute(relativeNormalVelocity, v1, v2, m2, -m1, elasticity);
                    }
                }
                else {
                    if (impulseScale) {
                        v1.minusVector(overlapVector, impulseScale);
                    }
                    if (edgeVector && isSeparating) {
                        v1.plusVector(relativeNormalVelocity, elasticity);
                    }
                }
            }
            else {
                if (body2.active) {
                    if (impulseScale) {
                        v2.plusVector(overlapVector, impulseScale);
                    }
                    if (edgeVector && isSeparating) {
                        v2.minusVector(relativeNormalVelocity, elasticity);
                    }
                }
                else {
                    return false;
                }
            }
            return edgeVector;
        }).forEach(collisionInfo => {
            // TODO: solve rotation
            const { body1, body2 } = collisionInfo;
            body1.contact.add(body2);
            body2.contact.add(body1);
            const { velocity: v1 } = body1, { velocity: v2 } = body2, friction = _min(body1.friction, body2.friction), staticFriction = _min(body1.staticFriction, body2.staticFriction);
            if (!staticFriction) {
                return;
            }
            const { overlap, edgeVector, relativeVelocity } = collisionInfo, relativeEdgeSpeed = Vector.project(relativeVelocity, edgeVector), absRelativeEdgeSpeed = _abs(relativeEdgeSpeed);
            if (body1.active) {
                if (body2.active) {
                    if (absRelativeEdgeSpeed < maxStaticSpeed &&
                        overlap * staticFriction >= absRelativeEdgeSpeed ||
                        overlap * friction >= absRelativeEdgeSpeed) {
                        const relativeEdgeVelocity = Vector.projectVector(relativeVelocity, edgeVector).shrink(2);
                        v1.plusVector(relativeEdgeVelocity);
                        v2.minusVector(relativeEdgeVelocity);
                    }
                    else if (friction) {
                        Vector.distribute(edgeVector, v1, v2, body2.mass, -body1.mass, overlap * friction * _sign(relativeEdgeSpeed));
                    }
                }
                else {
                    if (absRelativeEdgeSpeed < maxStaticSpeed &&
                        overlap * staticFriction >= absRelativeEdgeSpeed ||
                        overlap * friction >= absRelativeEdgeSpeed) {
                        v1.plusVector(edgeVector, relativeEdgeSpeed);
                    }
                    else if (friction) {
                        v1.plusVector(edgeVector, overlap * friction * _sign(relativeEdgeSpeed));
                    }
                }
            }
            else {
                if (absRelativeEdgeSpeed < maxStaticSpeed &&
                    overlap * staticFriction >= absRelativeEdgeSpeed ||
                    overlap * friction >= absRelativeEdgeSpeed) {
                    v2.minusVector(edgeVector, relativeEdgeSpeed);
                }
                else if (friction) {
                    v2.minusVector(edgeVector, overlap * friction * _sign(relativeEdgeSpeed));
                }
            }
        });
    },
    find(bodies, checker) {
        const results = new Array(), bodiesCount = bodies.length;
        for (let i = 0; i < bodiesCount; i++) {
            const body1 = bodies[i], { category: category1, velocity: velocity1 } = body1;
            for (let j = i + 1; j < bodiesCount; j++) {
                const body2 = bodies[j];
                if (!(category1 & body2.collisionFilter && body1.collisionFilter & body2.category)) {
                    continue;
                }
                const collisionInfo = checker(body1, body2);
                if (!collisionInfo) {
                    continue;
                }
                const { overlapVector } = collisionInfo;
                results.push(_assign(collisionInfo, {
                    body1, body2,
                    edgeVector: overlapVector.isZero() ?
                        _null :
                        overlapVector.clone().turn().normalize(),
                    relativeVelocity: Vector.minus(body2.velocity, velocity1)
                }));
            }
        }
        return results;
    },
    Checker: {
        AABB(body1, body2) {
            const { bounds: bounds1 } = body1, { bounds: bounds2 } = body2;
            let delta = bounds1.right - bounds2.left, min = delta, x = delta, y = 0;
            if (delta < 0) {
                return _null;
            }
            delta = bounds1.left - bounds2.right;
            if (delta > 0) {
                return _null;
            }
            else if (-delta < min) {
                min = -delta;
                x = delta;
            }
            delta = bounds1.bottom - bounds2.top;
            if (delta < 0) {
                return _null;
            }
            else if (delta < min) {
                min = delta;
                x = 0;
                y = delta;
            }
            delta = bounds1.top - bounds2.bottom;
            if (delta > 0) {
                return _null;
            }
            else if (-delta < min) {
                min = -delta;
                x = 0;
                y = delta;
            }
            return {
                overlap: min,
                overlapVector: Vector.of(x, y)
            };
        },
        SAT(body1, body2) {
            const { position: position1 } = body1;
            if (body2.isCircle && !body1.isCircle) {
                const result = Collision.Checker.SAT(body2, body1);
                if (result) {
                    result.overlapVector.reverse();
                }
                return result;
            }
            const normals = body1.normals.concat(body2.normals, Vector.minus(position1, body2.getClosest(position1)).normalize());
            if (!body1.isCircle) {
                const { position: position2 } = body2;
                normals.push(Vector.minus(position2, body1.getClosest(position2)).normalize());
            }
            let minDirection = _null, minOverlap = _Infinity;
            for (let i = 0; i < normals.length; i++) {
                const direction = normals[i], projection1 = body1.project(direction), projection2 = body2.project(direction);
                if (projection1.min > projection2.max || projection1.max < projection2.min) {
                    return _null;
                }
                const overlap1 = projection1.max - projection2.min, overlap2 = projection1.min - projection2.max;
                if (overlap1 < -overlap2) {
                    if (overlap1 < minOverlap) {
                        minOverlap = overlap1;
                        minDirection = direction;
                    }
                }
                else {
                    if (-overlap2 < minOverlap) {
                        minOverlap = -overlap2;
                        minDirection = direction.reverse();
                    }
                }
            }
            return minDirection && {
                overlap: minOverlap,
                overlapVector: minDirection.clone().scale(minOverlap)
            };
        },
        Distance(body1, body2) {
            const offset = Vector.minus(body2.position, body1.position), delta = body1.radius + body2.radius - offset.getNorm();
            return delta < 0 ? _null : {
                overlap: delta,
                overlapVector: offset.setNorm(delta)
            };
        },
        Smart(body1, body2) {
            if (body1.bounds.overlaps(body2.bounds)) {
                return (body1.isCircle && body2.isCircle) ?
                    Collision.Checker.Distance(body1, body2) :
                    Collision.Checker.SAT(body1, body2);
            }
            else {
                return _null;
            }
        },
    },
};

const Style = {
    Common: {
        defaults: {
            shadowColor: Utils.Const.TRANSPARENT,
            shadowBlur: 0,
            shadowOffsetX: 0,
            shadowOffsetY: 0,
            opacity: 1,
        },
        apply(renderer, style) {
            const { context, ratio } = renderer;
            context.globalAlpha = style.opacity;
            context.shadowColor = style.shadowColor;
            context.shadowBlur = style.shadowBlur * ratio;
            context.shadowOffsetX = style.shadowOffsetX * ratio;
            context.shadowOffsetY = style.shadowOffsetY * ratio;
        },
    },
    Fill: {
        defaults: {
            fillStyle: _null,
        },
        apply(renderer, style) {
            if (style.fillStyle) {
                const { context } = renderer;
                context.fillStyle = style.fillStyle;
            }
        }
    },
    Stroke: {
        defaults: {
            strokeStyle: '#000',
            lineWidth: 1,
            lineCap: 'butt',
            lineJoin: 'miter',
            miterLimit: 10,
            lineDash: _null,
            lineDashOffset: 0,
        },
        apply(renderer, style) {
            if (!style.strokeStyle) {
                return;
            }
            const { context } = renderer;
            context.strokeStyle = style.strokeStyle;
            context.lineWidth = style.lineWidth;
            context.lineCap = style.lineCap;
            context.lineJoin = style.lineJoin;
            if (style.lineDash) {
                context.setLineDash(style.lineDash);
                context.lineDashOffset = style.lineDashOffset;
            }
        },
    },
};

class Constraint {
    constructor(options) {
        _assign(this, Constraint.defaults, options);
        this.style = _assign({}, Constraint.defaultStyle, this.style);
        if (!this.originOffset) {
            this.originOffset = new Vector();
        }
        if (!this.targetOffset) {
            this.targetOffset = new Vector();
        }
        if (this.maxLength === _undefined) {
            const { origin, target } = this;
            this.maxLength = origin && target ?
                Vector.distance(origin.position || origin, target.position) :
                0;
        }
        if (this.minLength === _undefined) {
            this.minLength = 0;
        }
    }
    set length(length) {
        this.minLength = this.maxLength = length;
    }
    update(timeScale) {
        const { origin, target } = this;
        if (!target || !origin) {
            return;
        }
        const { minLength, maxLength } = this, elasticity = this.elasticity, { position: targetPosition } = target, originPosition = origin.position || origin, offsetVector = Vector.minus(targetPosition, originPosition)
            .minusVector(this.originOffset)
            .plusVector(this.targetOffset), offset = offsetVector.getNorm(), delta = offset > maxLength ? offset - maxLength :
            offset < minLength ? offset - minLength : 0;
        if (!delta) {
            return;
        }
        offsetVector.setNorm(delta);
        const originIsActive = originPosition !== origin && origin.active, { velocity: targetVelocity } = target, { stiffness } = this;
        if (originIsActive) {
            const { velocity: originVelocity } = origin;
            if (target.active) {
                const { mass: originMass } = origin, { mass: targetMass } = target, originScale = originMass / (originMass + targetMass), targetScale = 1 - originScale;
                origin.moveVector(offsetVector, originScale * stiffness);
                target.moveVector(offsetVector, -targetScale * stiffness);
                const relativeVelocity = Vector.minus(targetVelocity, originVelocity);
                if (Vector.dot(relativeVelocity, offsetVector)) {
                    const bounceVelocity = Vector.projectVector(relativeVelocity, offsetVector)
                        .plusVector(offsetVector)
                        .scale(elasticity);
                    originVelocity.plusVector(bounceVelocity, originScale);
                    targetVelocity.minusVector(bounceVelocity, targetScale);
                    // TODO: solve the rotations of origin & target here
                }
            }
            else {
                origin.moveVector(offsetVector);
                originVelocity.minusVector(Vector.projectVector(originVelocity, offsetVector), elasticity);
                // TODO: solve the rotation of origin here
            }
        }
        else {
            if (target.active || this.force) {
                target.move(-offsetVector.x, -offsetVector.y);
                targetVelocity.minusVector(Vector.projectVector(targetVelocity, offsetVector), elasticity);
                // TODO: solve the rotation of target here
            }
        }
    }
    render(renderer) {
        const { style, origin, target } = this;
        if (!style.strokeStyle || !origin || !target) {
            return;
        }
        const { originOffset, targetOffset } = this, { position: targetPosition } = target, originPosition = origin.position || origin, { context } = renderer;
        Style.Common.apply(renderer, style);
        Style.Stroke.apply(renderer, style);
        context.beginPath();
        context.moveTo(originPosition.x + originOffset.x, originPosition.y + originOffset.y);
        context.lineTo(targetPosition.x + targetOffset.x, targetPosition.y + targetOffset.y);
        context.stroke();
    }
}
Constraint.defaults = {
    origin: _null,
    target: _null,
    elasticity: 1,
    stiffness: 1,
    force: true,
    defer: true,
};
Constraint.defaultStyle = _assign({}, Style.Common.defaults, Style.Stroke.defaults);

const Sizing = {
    Full: (innerWidth, innerHeight, outerWidth, outerHeight, margin) => {
        const doubleMargin = margin * 2, width = outerWidth - doubleMargin, height = outerHeight - doubleMargin;
        return {
            width,
            height,
            styleWidth: width,
            styleHeight: height,
            left: margin,
            top: margin,
            scale: 1
        };
    },
    Center: (innerWidth, innerHeight, outerWidth, outerHeight, margin) => ({
        width: innerWidth,
        height: innerHeight,
        styleWidth: innerWidth,
        styleHeight: innerHeight,
        left: (outerWidth - innerWidth) / 2,
        top: (outerHeight - innerHeight) / 2,
        scale: 1
    }),
    Fit: (innerWidth, innerHeight, outerWidth, outerHeight, margin) => {
        const scale = outerWidth / outerHeight > innerWidth / innerHeight ?
            (outerHeight - margin * 2) / innerHeight :
            (outerWidth - margin * 2) / innerWidth, width = innerWidth * scale, height = innerHeight * scale;
        return {
            width: innerWidth,
            height: innerHeight,
            styleWidth: width,
            styleHeight: height,
            left: (outerWidth - width) / 2,
            top: (outerHeight - height) / 2,
            scale
        };
    },
    FixedWidth: (innerWidth, innerHeight, outerWidth, outerHeight, margin) => {
        const doubleMargin = margin * 2, styleWidth = outerWidth - doubleMargin, styleHeight = outerHeight - doubleMargin, scale = styleWidth / innerWidth;
        return {
            width: innerWidth,
            height: styleHeight / scale,
            styleWidth,
            styleHeight,
            left: margin,
            top: margin,
            scale
        };
    },
    FixedHeight: (innerWidth, innerHeight, outerWidth, outerHeight, margin) => {
        const doubleMargin = margin * 2, styleWidth = outerWidth - doubleMargin, styleHeight = outerHeight - doubleMargin, scale = styleHeight / innerHeight;
        return {
            width: styleWidth / scale,
            height: innerHeight,
            styleWidth,
            styleHeight,
            left: margin,
            top: margin,
            scale
        };
    },
    Fixed: (innerWidth, innerHeight, outerWidth, outerHeight, margin) => ((outerWidth / outerHeight > innerWidth / innerHeight ?
        Sizing.FixedHeight :
        Sizing.FixedWidth)(innerWidth, innerHeight, outerWidth, outerHeight, margin)),
};

class Layer {
    constructor(options) {
        this.canvas = _document.createElement('canvas');
        this.context = this.canvas.getContext('2d');
        _assign(this, Layer.defaults, options);
        if (!this.offset) {
            this.offset = new Vector();
        }
        if (this.objects) {
            this.expired = true;
        }
        else {
            this.expired = false;
            this.objects = [];
        }
        this.resize(this.width, this.height);
    }
    resize(width, height, ratio) {
        if (ratio) {
            this.ratio = ratio;
        }
        else {
            ratio = this.ratio;
        }
        const { canvas, offset, origin } = this, originX = this.originX = width * origin.x, originY = this.originY = height * origin.y;
        canvas.width = (this.width = width) * ratio;
        canvas.height = (this.height = height) * ratio;
        this.context.setTransform(ratio, 0, 0, ratio, (originX - offset.x) * ratio, (originY - offset.y) * ratio);
    }
    cache(immediate) {
        if (immediate) {
            this.expired = false;
            this.fill(this.background);
            this.objects.forEach(object => {
                object.render(this);
            });
        }
        else {
            this.expired = true;
        }
    }
    fill(color) {
        const { context, originX, originY, width, height } = this;
        if (color) {
            context.fillStyle = color;
            context.fillRect(-originX, -originY, width, height);
        }
        else {
            context.clearRect(-originX, -originY, width, height);
        }
    }
    update(timeScale) {
        if (this.active) {
            this.objects.filter(object => {
                if (object.update) {
                    if (object.defer) {
                        return true;
                    }
                    else {
                        object.update(timeScale);
                    }
                }
            }).forEach(object => {
                object.update(timeScale);
            });
        }
        if (this.expired) {
            this.cache(true);
        }
    }
    render(renderer) {
        Style.Common.apply(renderer, Style.Common.defaults);
        const { offset } = this;
        renderer.context.drawImage(this.canvas, offset.x - this.originX, offset.y - this.originY, this.width, this.height);
    }
}
Layer.defaults = {
    active: true,
    width: 480,
    height: 320,
    ratio: Utils.Const.IS_TOUCH_MODE ? _window.devicePixelRatio || 1 : 2,
    origin: Vector.of(.5, .5),
    background: _null,
};

class Renderer extends EventEmitter {
    constructor(options = Utils.Const.EMPTY_OBJECT) {
        super();
        this.settings = Utils.Const.EMPTY_OBJECT;
        this.bounds = new Bounds();
        this.resizeListener = Utils.debounce(this._resizeListener.bind(this));
        _assign(this, Renderer.defaults, options);
        let { canvas } = this;
        if (!canvas) {
            canvas = this.canvas || (this.canvas = _document.createElement('canvas'));
        }
        else if (canvas.parentNode) {
            this.parent = canvas.parentElement;
        }
        if (this.parent) {
            this.parent.appendChild(canvas);
        }
        this.context = canvas.getContext('2d', this.settings);
        const { resizeListener } = this;
        this.resizeEvents.forEach(event => {
            _window.addEventListener(event, resizeListener);
        });
        this._width = this.width;
        this._height = this.height;
        this._resize();
    }
    set resizeDelay(delay) {
        this.resizeListener.delay = delay;
    }
    get resizeDelay() {
        return this.resizeListener.delay;
    }
    _resize() {
        const { canvas, ratio, origin } = this, { style } = canvas;
        const { _width, _height } = this;
        let width = _width, height = _height, styleWidth = _width, styleHeight = _height;
        if (this.align) {
            const { parent } = this;
            if (parent) {
                const rect = parent.getBoundingClientRect(), size = this.sizing(_width, _height, rect.width, rect.height, this.margin);
                style.marginLeft = size.left + 'px';
                style.marginTop = size.top + 'px';
                width = this.width = size.width;
                height = this.height = size.height;
                styleWidth = size.styleWidth;
                styleHeight = size.styleHeight;
                this._offsetX = rect.left + size.left;
                this._offsetY = rect.top + size.top;
                this._scale = size.scale;
            }
        }
        else {
            this._scale = 1;
        }
        style.width = styleWidth + 'px';
        style.height = styleHeight + 'px';
        canvas.width = width * ratio;
        canvas.height = height * ratio;
        this.emit('resize');
        const originX = this.originX = width * origin.x, originY = this.originY = height * origin.y;
        this.context.setTransform(ratio, 0, 0, ratio, originX * ratio, originY * ratio);
        const { bounds } = this;
        bounds.top = -originY;
        bounds.right = -originX + width;
        bounds.bottom = -originY + height;
        bounds.left = -originX;
    }
    _resizeListener() {
        this._resize();
    }
    resize(width, height, ratio) {
        this._width = this.width = width;
        this._height = this.height = height;
        if (ratio) {
            this.ratio = ratio;
        }
        this._resize();
        return this;
    }
    fill(color) {
        const { context, bounds } = this, { left, top, width, height } = bounds;
        if (color) {
            context.fillStyle = color;
            context.fillRect(left, top, width, height);
        }
        else {
            context.clearRect(left, top, width, height);
        }
    }
    render(renderable) {
        const { context, restoration } = this;
        if (restoration) {
            context.save();
        }
        renderable.render(this);
        if (restoration) {
            context.restore();
        }
    }
    toInnerPosition(position) {
        const { _scale } = this;
        return Vector.of((position.x - this._offsetX) / _scale - this.originX, (position.y - this._offsetY) / _scale - this.originY);
    }
    toOuterPosition(position) {
        const { _scale } = this;
        return Vector.of((position.x + this.originX) * _scale + this._offsetX, (position.y + this.originY) * _scale + this._offsetY);
    }
    createLayer(options) {
        const { bounds } = this;
        return new Layer(_assign({
            width: bounds.width,
            height: bounds.height,
            ratio: this.ratio,
        }, options));
    }
}
Renderer.defaults = {
    settings: Utils.Const.EMPTY_OBJECT,
    width: Layer.defaults.width,
    height: Layer.defaults.height,
    margin: 0,
    ratio: Layer.defaults.ratio,
    origin: Layer.defaults.origin,
    parent: _document.body,
    align: true,
    sizing: Sizing.Fit,
    resizeEvents: ['resize', 'orientationchange'],
    resizeDelay: 100,
    restoration: false,
};

class Sprite {
    constructor(options = Utils.Const.EMPTY_OBJECT) {
        _assign(this, options);
        if (!this.position) {
            this.position = new Vector();
        }
        this.style = _assign({}, Sprite.defaultStyle, this.style);
    }
    static of(src, options) {
        return (new Sprite(options)).load(src);
    }
    load(src, onSuccess, onFailure) {
        const image = this.image = _document.createElement('img');
        image.src = src;
        if (onSuccess) {
            image.onload = onSuccess;
        }
        if (onFailure) {
            image.onabort = image.onerror = onFailure;
        }
        return this;
    }
    render(renderer) {
        if (!this.visible) {
            return;
        }
        const { image } = this;
        if (image) {
            const { position, rotation } = this, { width, height } = image, { context } = renderer, dstW = this.width || width, dstH = this.height || height;
            Style.Common.apply(renderer, this.style);
            context.rotate(rotation);
            context.drawImage(image, (this.clipX || 0) - dstW / 2, (this.clipY || 0) - dstH / 2, this.clipWidth || width, this.clipHeight || height, position.x, position.y, dstW, dstH);
            context.rotate(-rotation);
        }
    }
}
Sprite.defaults = {
    visible: true,
    rotation: 0,
};
Sprite.defaultStyle = _assign({}, Style.Common.defaults);

class Shape extends Body {
    constructor(options = Utils.Const.EMPTY_OBJECT) {
        super(_assign({}, Shape.defaults, options));
        this.style = _assign({}, Shape.defaultStyle, this.style);
        if (!this.attachments) {
            this.attachments = [];
        }
    }
    static applyStyle(renderer, style) {
        Style.Common.apply(renderer, style);
        Style.Fill.apply(renderer, style);
        Style.Stroke.apply(renderer, style);
    }
    render(renderer) {
        if (!this.visible) {
            return;
        }
        const { style, fillFirst, position, texture, attachments } = this, { fillStyle } = style, { context } = renderer, { TRANSPARENT } = Utils.Const;
        context.translate(position.x, position.y);
        if (texture) {
            const { rotation } = this;
            context.rotate(rotation);
            texture.render(renderer);
            context.rotate(-rotation);
        }
        else {
            if (this.preferShadow && !style.shadowBlur && style.shadowColor !== TRANSPARENT) {
                context.shadowColor = TRANSPARENT;
                const { shadowOffsetX, shadowOffsetY } = style;
                context.translate(shadowOffsetX, shadowOffsetY);
                context.beginPath();
                this.path(context);
                if (this.closePath) {
                    context.closePath();
                }
                context.fillStyle = style.shadowColor;
                context.fill();
                context.translate(-shadowOffsetX, -shadowOffsetY);
                if (fillStyle) {
                    context.fillStyle = fillStyle;
                }
            }
            Shape.applyStyle(renderer, style);
            context.beginPath();
            this.path(context);
            if (this.closePath) {
                context.closePath();
            }
            if (fillFirst && fillStyle) {
                context.fill();
                context.shadowColor = TRANSPARENT;
            }
            if (style.strokeStyle) {
                context.stroke();
                context.shadowColor = TRANSPARENT;
            }
            if (!fillFirst && fillStyle) {
                context.fill();
            }
            if (attachments.length) {
                attachments.forEach(attachment => {
                    attachment.render(renderer);
                });
            }
        }
        context.translate(-position.x, -position.y);
    }
}
Shape.defaults = {
    visible: true,
    fillFirst: true,
    closePath: true,
    preferShadow: false,
    texture: _null,
};
Shape.defaultStyle = _assign({}, Style.Common.defaults, Style.Stroke.defaults, { fillStyle: _null });

class Text {
    constructor(options) {
        _assign(this, Text.defaults, options);
        if (!this.position) {
            this.position = new Vector();
        }
        this.style = _assign({}, Text.defaultStyle, this.style);
    }
    static applyStyle(renderer, textStyle) {
        Shape.applyStyle(renderer, textStyle);
        const { context } = renderer;
        context.font = textStyle.font;
        context.textAlign = textStyle.textAlign;
        context.textBaseline = textStyle.textBaseline;
        context.direction = textStyle.direction;
    }
    render(renderer) {
        const { style, fillFirst, content, position, rotation } = this, { x, y } = position, { fillStyle } = style, { context } = renderer, { TRANSPARENT } = Utils.Const;
        Text.applyStyle(renderer, style);
        context.translate(x, y);
        context.rotate(rotation);
        if (this.preferShadow && !style.shadowBlur && style.shadowColor !== TRANSPARENT) {
            context.shadowColor = TRANSPARENT;
            context.fillStyle = style.shadowColor;
            context.fillText(content, style.shadowOffsetX, style.shadowOffsetY);
            if (fillStyle) {
                context.fillStyle = fillStyle;
            }
        }
        if (fillFirst && fillStyle) {
            context.fillText(content, 0, 0);
            context.shadowColor = TRANSPARENT;
        }
        if (style.strokeStyle) {
            context.strokeText(content, 0, 0);
            context.shadowColor = TRANSPARENT;
        }
        if (!fillFirst && fillStyle) {
            context.fillText(content, 0, 0);
        }
        context.rotate(-rotation);
        context.translate(-x, -y);
    }
}
Text.defaults = {
    visible: true,
    content: '',
    preferShadow: false,
    rotation: 0,
};
Text.defaultStyle = _assign({}, Shape.defaultStyle, {
    font: '16px Consolas',
    textAlign: 'center',
    textBaseline: 'middle',
    direction: 'inherit',
});

class Paragraph {
    constructor(options) {
        _assign(this, Paragraph.defaults, options);
        if (!this.position) {
            this.position = new Vector();
        }
        this.style = _assign({}, Paragraph.defaultStyle, this.style);
    }
    render(renderer) {
        const { style, fillFirst, lines, lineHeight, indent, position, rotation } = this, { fillStyle, strokeStyle, shadowColor, shadowOffsetX, shadowOffsetY } = style, { x, y } = position, { context } = renderer, { TRANSPARENT } = Utils.Const, preferShadow = this.preferShadow && !style.shadowBlur && style.shadowColor !== TRANSPARENT;
        Text.applyStyle(renderer, style);
        if (preferShadow) {
            context.shadowColor = TRANSPARENT;
        }
        context.translate(x, y);
        context.rotate(rotation);
        let dx = 0, dy = 0;
        lines.forEach(line => {
            if (preferShadow) {
                context.fillStyle = style.shadowColor;
                context.fillText(line, dx + shadowOffsetX, dy + shadowOffsetY);
                if (fillStyle) {
                    context.fillStyle = fillStyle;
                }
            }
            if (fillFirst && fillStyle) {
                context.fillText(line, dx, dy);
                context.shadowColor = TRANSPARENT;
            }
            if (strokeStyle) {
                context.strokeText(line, dx, dy);
                context.shadowColor = TRANSPARENT;
            }
            if (!fillFirst && fillStyle) {
                context.fillText(line, dx, dy);
            }
            context.shadowColor = shadowColor;
            dx += indent;
            dy += lineHeight;
        });
        context.rotate(-rotation);
        context.translate(-x, -y);
    }
}
Paragraph.defaults = {
    visible: true,
    lines: [],
    indent: 0,
    lineHeight: 20,
    preferShadow: true,
    rotation: 0,
};
Paragraph.defaultStyle = _assign({}, Text.defaultStyle);

class Rectangle extends Shape {
    constructor(options = Utils.Const.EMPTY_OBJECT) {
        super(_assign({}, Rectangle.defaults, options));
        this.updateBounds();
        const { rotation } = this, cos = _cos(rotation), sin = _sin(rotation);
        this.normals = [
            Vector.of(cos, sin),
            Vector.of(-sin, cos)
        ];
        this.updateSize();
    }
    _scale(scaleX, scaleY, origin) {
        this.width *= scaleX;
        this.height *= scaleY;
        const { position: { x: x0, y: y0 }, bounds } = this, x = origin ? origin.x + (x0 - origin.x) * scaleX : x0, y = origin ? origin.y + (y0 - origin.y) * scaleY : y0, dx = (bounds.right - x) * scaleX, dy = (bounds.bottom - y) * scaleY;
        bounds.left = x - dx;
        bounds.right = x + dx;
        bounds.top = y - dy;
        bounds.bottom = y + dy;
    }
    _rotate(rotation, origin) {
        const { normals } = this;
        normals[0].rotate(rotation);
        normals[1].rotate(rotation);
        this.updateBounds();
    }
    updateSize(width, height) {
        if (width === _undefined) {
            width = this.width;
        }
        else {
            this.width = width;
        }
        if (height === _undefined) {
            height = this.height;
        }
        else {
            this.height = height;
        }
        this._setArea(width * height);
        this.updateBounds();
    }
    updateBounds() {
        const { bounds, rotation, radius, width, height, position } = this, { x, y } = position;
        let dx0 = width / 2, dy0 = height / 2;
        if (radius > 0) {
            dx0 -= radius;
            dy0 -= radius;
        }
        const cos = _cos(rotation), sin = _sin(rotation), dx1 = dx0 * cos - dy0 * sin, dy1 = dx0 * sin + dy0 * cos, dx2 = dx0 * cos - -dy0 * sin, dy2 = dx0 * sin + -dy0 * cos;
        bounds.updateVertices(Vertices.fromArray([
            x + dx1, y + dy1,
            x + dx2, y + dy2,
            x - dx1, y - dy1,
            x - dx2, y - dy2
        ]));
        if (radius > 0) {
            bounds.grow(radius);
        }
    }
    getClosest(target) {
        const { rotation, radius, width, height, position } = this, { x, y } = position;
        let dx0 = width / 2, dy0 = height / 2;
        if (radius > 0) {
            dx0 -= radius;
            dy0 -= radius;
        }
        const cos = _cos(rotation), sin = _sin(rotation), dx1 = dx0 * cos - dy0 * sin, dy1 = dx0 * sin + dy0 * cos, dx2 = dx0 * cos - -dy0 * sin, dy2 = dx0 * sin + -dy0 * cos, closest = Vertices.findClosest(target, Vertices.fromArray([
            x + dx1, y + dy1,
            x + dx2, y + dy2,
            x - dx1, y - dy1,
            x - dx2, y - dy2
        ]));
        if (radius > 0) {
            closest.grow(radius);
        }
        return closest;
    }
    project(direction) {
        const { radius } = this, angle = this.rotation - direction.getAngle(), cos = _cos(angle), sin = _sin(angle);
        let x0 = this.width / 2, y0 = this.height / 2;
        if (radius > 0) {
            x0 -= radius;
            y0 -= radius;
        }
        let halfLength = _max(_abs(x0 * cos - y0 * sin), _abs(x0 * cos + y0 * sin));
        if (radius > 0) {
            halfLength += radius;
        }
        const positionProjection = Vector.project(this.position, direction);
        return {
            min: positionProjection - halfLength,
            max: positionProjection + halfLength
        };
    }
    path(context) {
        const { width, height, radius, rotation } = this;
        context.rotate(rotation);
        if (radius > 0) {
            const halfWidth = width / 2, halfHeight = height / 2, innerHalfWidth = halfWidth - radius, innerHalfHeight = halfHeight - radius;
            context.moveTo(-innerHalfWidth, -halfHeight);
            context.arcTo(halfWidth, -halfHeight, halfWidth, innerHalfHeight, radius);
            context.arcTo(halfWidth, halfHeight, -innerHalfWidth, halfHeight, radius);
            context.arcTo(-halfWidth, halfHeight, -halfWidth, -innerHalfHeight, radius);
            context.arcTo(-halfWidth, -halfHeight, -innerHalfWidth, -halfHeight, radius);
        }
        else {
            context.rect(-width / 2, -height / 2, width, height);
        }
        context.rotate(-rotation);
    }
}
Rectangle.defaults = {
    width: 20,
    height: 20,
    radius: 0,
};

class Circle extends Shape {
    constructor(options = Utils.Const.EMPTY_OBJECT) {
        super(_assign({}, Circle.defaults, options));
        this.isCircle = true;
        if (this.radius) {
            this.updateRadius();
        }
    }
    updateRadius(radius) {
        if (radius === _undefined) {
            radius = this.radius;
        }
        else {
            this.radius = radius;
        }
        this._setArea(radius * radius * _PI);
        this.updateBounds();
    }
    updateBounds() {
        const { bounds, position: { x, y }, radius, rotation } = this, cos = _cos(rotation), sin = -_sin(rotation), a = radius * this.scaleX, b = radius * this.scaleY, halfWidth = _sqrt(Utils.quadraticSum(a * cos, b * sin)), halfHeight = _sqrt(Utils.quadraticSum(b * cos, a * sin));
        bounds.left = x - halfWidth;
        bounds.right = x + halfWidth;
        bounds.top = y - halfHeight;
        bounds.bottom = y + halfHeight;
    }
    getClosest(target) {
        const { position } = this, deltaVector = Vector.minus(target, position), angle = deltaVector.getAngle();
        return deltaVector.setNorm(this.radius * _sqrt(Utils.quadraticSum(_cos(angle) * this.scaleX, _sin(angle) * this.scaleY))).plusVector(position);
    }
    project(direction) {
        const positionProjection = Vector.project(this.position, direction), angle = this.rotation - direction.getAngle(), halfLength = this.radius * _sqrt(Utils.quadraticSum(_cos(angle) * this.scaleX, _sin(angle) * this.scaleY));
        return {
            min: positionProjection - halfLength,
            max: positionProjection + halfLength
        };
    }
    path(context) {
        const { rotation, scaleX, scaleY } = this;
        context.rotate(rotation);
        context.scale(scaleX, scaleY);
        context.arc(0, 0, this.radius, 0, Utils.Const.DOUBLE_PI);
        context.rotate(-rotation);
        context.scale(1 / scaleX, 1 / scaleY);
    }
}
Circle.defaults = {
    radius: 1,
};

class Polygon extends Shape {
    constructor(options = Utils.Const.EMPTY_OBJECT) {
        super(_assign({}, Polygon.defaults, options));
        if (this.vertices) {
            this.updateVertices();
        }
        else {
            this.vertices = [];
        }
    }
    updateVertices(vertices) {
        if (vertices) {
            this.vertices = vertices;
        }
        else {
            vertices = this.vertices;
        }
        if (vertices.length > 1) {
            const { clockwise, adjustment } = this, { normalPrecision } = Body, areas = new Array(), normals = new Array(), centers = new Array(), tangents = new _Set();
            let totalArea = 0;
            vertices.reduce((vertex1, vertex2) => {
                const area = Vector.cross(vertex1, vertex2) / 2;
                totalArea += area;
                const normal = Vector.minus(vertex2, vertex1).turn(clockwise), tangent = (normal.y / normal.x).toFixed(normalPrecision);
                if (!tangents.has(tangent)) {
                    tangents.add(tangent);
                    normals.push(normal.normalize());
                }
                if (adjustment) {
                    centers.push(Vector.of((vertex1.x + vertex2.x) / 3, (vertex1.y + vertex2.y) / 3));
                    areas.push(area);
                }
                return vertex2;
            }, vertices[vertices.length - 1]);
            if (adjustment) {
                const offset = Vector.plus(centers.map((center, i) => center.scale(areas[i] / totalArea)));
                vertices.forEach(vertex => {
                    vertex.minusVector(offset);
                });
            }
            if (!clockwise) {
                totalArea = -totalArea;
            }
            this.normals = normals;
            this._setArea(totalArea);
        }
        else {
            this._setArea(0);
        }
        this.updateBounds();
        return this;
    }
    _scale(scaleX, scaleY, origin) {
        this.vertices.forEach(vertex => {
            vertex.scale(scaleX, scaleY, origin);
        });
    }
    _rotate(rotation, origin) {
        this.vertices.forEach(vertex => {
            vertex.rotate(rotation, origin);
        });
        this.normals.forEach(normal => {
            normal.rotate(rotation, origin);
        });
    }
    updateBounds() {
        const { vertices, bounds, position } = this;
        if (vertices.length) {
            bounds.updateVertices(vertices);
            bounds.moveVector(position);
        }
        else {
            bounds.left = bounds.right = position.x;
            bounds.top = bounds.bottom = position.y;
        }
    }
    getClosest(target) {
        return Vertices.findClosest(target, this.vertices);
    }
    project(direction) {
        const { vertices } = this;
        let min = 0, max = 0, projection;
        if (vertices.length) {
            vertices.forEach((vertex, i) => {
                projection = Vector.project(vertex, direction);
                if (i > 0) {
                    if (projection < min) {
                        min = projection;
                    }
                    else if (projection > max) {
                        max = projection;
                    }
                }
                else {
                    min = max = projection;
                }
            });
        }
        const positionProjection = Vector.project(this.position, direction);
        return {
            min: min + positionProjection,
            max: max + positionProjection
        };
    }
    path(context) {
        this.vertices.forEach((vertex, i) => {
            if (i > 0) {
                context.lineTo(vertex.x, vertex.y);
            }
            else {
                context.moveTo(vertex.x, vertex.y);
            }
        });
    }
}
Polygon.defaults = {
    clockwise: true,
    adjustment: true,
};

class Line extends Shape {
    constructor(options = Utils.Const.EMPTY_OBJECT) {
        super(_assign({}, Line.defaults, options));
        if (!this.start) {
            this.start = new Vector();
        }
        if (!this.end) {
            this.end = new Vector();
        }
        this.updateVertices();
    }
    _scale(scaleX, scaleY, origin) {
        this.start.scale(scaleX, scaleY, origin);
        this.end.scale(scaleX, scaleY, origin);
        this.updateVertices();
    }
    _rotate(rotation, origin) {
        this.start.rotate(rotation, origin);
        this.end.rotate(rotation, origin);
        this.updateVertices();
    }
    updateVertices(start, end) {
        if (start) {
            this.start = start;
        }
        if (end) {
            this.end = end;
        }
        const deltaVector = Vector.minus(this.end, this.start);
        this.normals = [
            deltaVector,
            Vector.of(-deltaVector.y, deltaVector.x)
        ];
        this._setArea(deltaVector.getNorm());
        this.updateBounds();
    }
    updateBounds() {
        const { start: { x: x1, y: y1 }, end: { x: x2, y: y2 }, bounds } = this;
        if (x1 > x2) {
            bounds.left = x2;
            bounds.right = x1;
        }
        else {
            bounds.left = x1;
            bounds.right = x2;
        }
        if (y1 > y2) {
            bounds.top = y2;
            bounds.bottom = y1;
        }
        else {
            bounds.top = y1;
            bounds.bottom = y2;
        }
    }
    getClosest(target) {
        const { start, end } = this;
        return Utils.quadraticSum(start.x - target.x, start.y - target.y) <
            Utils.quadraticSum(end.x - target.x, end.y - target.y) ?
            start : end;
    }
    project(direction) {
        const { start, end } = this, startProjection = Vector.project(start, direction), endProjection = Vector.project(end, direction);
        return startProjection < endProjection ?
            { min: startProjection, max: endProjection } :
            { min: endProjection, max: startProjection };
    }
    path(context) {
        const { start, end } = this;
        context.moveTo(start.x, start.y);
        context.lineTo(end.x, end.y);
    }
}
Line.defaults = {
    closePath: false,
};

class Particles {
    constructor(options) {
        this._items = [];
        this._deadlines = new _Map();
        this._legacies = new _Map();
        this.spawn = Utils.throttle(this._spawn);
        _assign(this, Particles.defaults, options);
        if (!this.offset) {
            this.offset = new Vector();
        }
        if (!this.items) {
            this.items = [];
        }
    }
    set life(value) {
        this.minLife = this.maxLife = value;
    }
    set count(value) {
        this.minCount = this.maxCount = value;
    }
    _spawn() {
        const { pool } = this;
        if (!pool) {
            return;
        }
        const { items, _deadlines, initializer } = this, particle = pool.get(), life = Utils.mix(this.minLife, this.maxLife, _Math.random());
        _deadlines.set(particle, _now() + life);
        if (initializer) {
            this._legacies.set(particle, initializer(particle, life));
        }
        this._items.push(particle);
        if (this.sequential) {
            items.push(particle);
        }
        else {
            Utils.insert(items, _Math.round(items.length * _Math.random()), particle);
        }
    }
    set threshold(value) {
        this.spawn.threshold = value;
    }
    get threshold() {
        return this.spawn.threshold;
    }
    clear() {
        const { pool, cleaner, _legacies } = this;
        this.items.forEach(item => {
            if (pool) {
                pool.add(item);
            }
            if (cleaner) {
                cleaner(item, _legacies.get(item));
            }
        });
        this.items = [];
        this._items.length = 0;
        this._deadlines.clear();
        this._legacies.clear();
    }
    _remove(item) {
        const { _deadlines, pool, cleaner, _legacies, _items } = this;
        Utils.removeIndex(_items, _items.indexOf(item));
        _deadlines.delete(item);
        if (pool) {
            pool.add(item);
        }
        if (cleaner) {
            cleaner(item, _legacies.get(item));
        }
        _legacies.delete(item);
    }
    update(timeScale) {
        if (!this.active) {
            return;
        }
        const { _deadlines } = this, now = _now();
        this.items = this.items.filter(item => {
            if (_deadlines.get(item) <= now) {
                return this._remove(item);
            }
            if (item.update) {
                item.update(timeScale);
            }
            return true;
        });
        if (this.loop) {
            this.spawn();
        }
        const currentCount = this.items.length;
        if (currentCount > this.maxCount) {
            const { items } = this;
            if (this.sequential) {
                for (let i = this.maxCount; i < currentCount; i++) {
                    this._remove(items.shift());
                }
            }
            else {
                const { _items } = this;
                for (let i = this.maxCount; i < currentCount; i++) {
                    const item = _items[0];
                    this._remove(item);
                    Utils.removeIndex(items, items.indexOf(item));
                }
            }
        }
        else if (currentCount < this.minCount) {
            for (let i = currentCount; i < this.minCount; i++) {
                this._spawn();
            }
        }
    }
    render(renderer) {
        if (!this.visible) {
            return;
        }
        const { items, offset } = this, { context } = renderer;
        context.translate(offset.x, offset.y);
        items.forEach(item => {
            item.render(renderer);
        });
        context.translate(-offset.x, -offset.y);
    }
}
Particles.defaults = {
    visible: true,
    active: true,
    pool: _null,
    minCount: 0,
    maxCount: _Infinity,
    minLife: _Infinity,
    maxLife: _Infinity,
    initializer: _null,
    cleaner: _null,
    sequential: false,
    loop: true,
};

class Timer extends EventEmitter {
    constructor(options) {
        super();
        this.lastFrameDelay = 0;
        this.lastFrameDuration = 0;
        this.isRunning = false;
        this._timeoutId = 0;
        this._timeoutCallbacks = new Map();
        this._intervalId = 0;
        this._intervalCallbacks = new Map();
        _assign(this, Timer.defaults, options);
        this._tick = this._tick.bind(this);
    }
    _requestTick(delay) {
        this._timer = (this._usedRAF = this.delay <= Timer.RAFThreshold && this.allowRAF) ?
            _window.requestAnimationFrame(this._tick) :
            _setTimeout(this._tick, _max(delay, 0));
    }
    _tick() {
        const startTime = _now(), deltaTime = this.lastFrameDelay = startTime - this._lastTickTime, { _timer } = this;
        this._lastTickTime = startTime;
        const { _timeoutCallbacks } = this;
        _timeoutCallbacks.forEach((record, id) => {
            if (startTime >= record[1]) {
                record[0].apply(_undefined, record[2]);
                _timeoutCallbacks.delete(id);
            }
        });
        this._intervalCallbacks.forEach(record => {
            if (startTime >= record[1]) {
                record[0].apply(_undefined, record[3]);
                while (startTime >= record[1]) {
                    record[1] += record[2];
                }
            }
        });
        this.emit('tick', deltaTime);
        if (!this.isRunning) {
            return;
        }
        const duration = this.lastFrameDuration = _now() - startTime;
        if (this._timer === _timer) {
            this._requestTick(_max(this.delay - (this.fixDelay ? duration : 0), 0));
        }
    }
    start() {
        this.isRunning = true;
        const now = (this._lastTickTime = _now()), { _lastStopTime } = this;
        if (_lastStopTime !== _undefined) {
            const delay = now - _lastStopTime;
            this._timeoutCallbacks.forEach(record => {
                record[1] += delay;
            });
            this._intervalCallbacks.forEach(record => {
                record[1] += delay;
            });
        }
        this._tick();
    }
    stop() {
        if (!this.isRunning) {
            return;
        }
        this.isRunning = false;
        this._lastStopTime = _now();
        const { _timer } = this;
        if (_timer !== _undefined) {
            (this._usedRAF ? _window.cancelAnimationFrame : _clearTimeout)(_timer);
            this._timer = _undefined;
        }
    }
    reschedule(delay) {
        this.delay = delay;
        if (delay < 0) {
            return this.stop();
        }
        if (this.isRunning) {
            this.stop();
            this.isRunning = true;
            const now = _now();
            this._requestTick(_max(delay - (now - this._lastTickTime), 0));
            this._lastTickTime = now;
        }
    }
    setTimeout(callback, timeout, ...args) {
        const id = this._timeoutId++;
        this._timeoutCallbacks.set(id, [callback, _now() + timeout, args]);
        return id;
    }
    clearTimeout(id) {
        this._timeoutCallbacks.delete(id);
    }
    setInterval(callback, interval, ...args) {
        const id = this._intervalId++;
        this._intervalCallbacks.set(id, [callback, _now() + interval, interval, args]);
        return id;
    }
    clearInterval(id) {
        this._intervalCallbacks.delete(id);
    }
}
Timer.RAFThreshold = 1000 / 60;
Timer.defaults = {
    delay: Timer.RAFThreshold,
    allowRAF: true,
    fixDelay: true,
};

class Pointer extends EventEmitter {
    constructor(options) {
        super();
        this.isCircle = true;
        this.position = new Vector();
        this.bounds = new Bounds();
        this.normals = [];
        this.active = true;
        this.isHolding = false;
        this.startTimeStamps = new _Map();
        this.positions = new _Map();
        _assign(this, Pointer.defaults, options);
        const { target } = this;
        if (this.isTouchMode) {
            target.addEventListener('touchstart', this._startListener = ((event) => {
                const touchPoint = event.changedTouches[0];
                this._start(touchPoint.identifier, Vector.of(touchPoint.clientX, touchPoint.clientY), event);
            }));
            target.addEventListener('touchmove', this._moveListener = ((event) => {
                const touchPoint = event.changedTouches[0];
                this._move(touchPoint.identifier, Vector.of(touchPoint.clientX, touchPoint.clientY), event);
            }));
            target.addEventListener('touchend', this._endListener = ((event) => {
                const touchPoint = event.changedTouches[0];
                this._end(touchPoint.identifier, Vector.of(touchPoint.clientX, touchPoint.clientY), event);
            }));
        }
        else {
            target.addEventListener('mousedown', this._startListener = ((event) => {
                this._start(-1, Vector.of(event.clientX, event.clientY), event);
            }));
            target.addEventListener('mousemove', this._moveListener = ((event) => {
                this._move(-1, Vector.of(event.clientX, event.clientY), event);
            }));
            target.addEventListener('mouseup', this._endListener = ((event) => {
                this._end(-1, Vector.of(event.clientX, event.clientY), event);
            }));
        }
    }
    _setPosition(id, rawPosition, endFlag) {
        const { bounds, radius, positions, pretransform, transform } = this, tempPosition = pretransform ? pretransform(rawPosition) : rawPosition, newPosition = transform ? transform(tempPosition) : tempPosition;
        this.position.setVector(newPosition);
        if (endFlag) {
            positions.delete(id);
        }
        else {
            positions.set(id, newPosition);
        }
        bounds.left = newPosition.x - radius;
        bounds.right = newPosition.x + radius;
        bounds.top = newPosition.y - radius;
        bounds.bottom = newPosition.y + radius;
    }
    _start(id, rawPosition, event) {
        this.isHolding = true;
        this.startTimeStamps.set(id, event.timeStamp);
        this._setPosition(id, rawPosition);
        this.emit('start', this.position, id, event);
    }
    _move(id, rawPosition, event) {
        if (this.holdOnly && !this.isHolding) {
            return;
        }
        this._setPosition(id, rawPosition);
        this.emit('move', this.position, id, event);
    }
    _end(id, rawPosition, event) {
        this._setPosition(id, rawPosition, true);
        this.isHolding = false;
        const { startTimeStamps } = this, startTimeStamp = startTimeStamps.get(id);
        startTimeStamps.delete(id);
        this.emit('end', this.position, id, event);
        if (event.timeStamp - startTimeStamp <= this.clickThreshold) {
            this.emit('click', this.position, id, event);
        }
    }
    getClosest(target) {
        return Vector.minus(target, this.position).setNorm(this.radius);
    }
    project(direction) {
        const { radius } = this, positionProjection = Vector.project(this.position, direction);
        return {
            min: positionProjection - radius,
            max: positionProjection + radius
        };
    }
    destroy() {
        if (!this.active) {
            return;
        }
        this.active = false;
        const { target } = this;
        if (this.isTouchMode) {
            target.removeEventListener('touchstart', this._startListener);
            target.removeEventListener('touchmove', this._moveListener);
            target.removeEventListener('touchend', this._endListener);
        }
        else {
            target.removeEventListener('mousedown', this._startListener);
            target.removeEventListener('mousemove', this._moveListener);
            target.removeEventListener('mouseup', this._endListener);
        }
    }
}
Pointer.defaults = {
    target: _window,
    pretransform: _null,
    transform: _null,
    isTouchMode: Utils.Const.IS_TOUCH_MODE,
    clickThreshold: 1000,
};
Pointer.defaults.radius = Pointer.defaults.isTouchMode ? 5 : 0;

class Camera {
    constructor(options) {
        _assign(this, Camera.defaults, options);
        if (!this.position) {
            this.position = new Vector();
        }
        if (!this.scale) {
            this.scale = new Vector(1, 1);
        }
    }
    applyTo(context) {
        const { position, scale } = this, { x: scaleX, y: scaleY } = scale;
        context.transform(scaleX, 0, 0, scaleY, -position.x * scaleX, -position.y * scaleY);
        context.rotate(this.rotation);
    }
    restore(context) {
        const { position, scale } = this;
        context.rotate(-this.rotation);
        context.transform(1 / scale.x, 0, 0, 1 / scale.y, position.x, position.y);
    }
    toViewPosition(position) {
        const { position: pos, scale, rotation } = this, cos = _cos(rotation), sin = -_sin(rotation), tx = (position.x + pos.x) / scale.x, ty = (position.y + pos.y) / scale.y;
        return Vector.of(tx * cos - ty * sin, tx * sin + ty * cos);
    }
    toGlobalPosition(position) {
        const { position: pos, scale, rotation } = this, cos = _cos(rotation), sin = _sin(rotation), tx = position.x * scale.x - pos.x, ty = position.y * scale.y - pos.y;
        return Vector.of(tx * cos - ty * sin, tx * sin + ty * cos);
    }
}
Camera.defaults = {
    rotation: 0,
};

class Scene extends EventEmitter {
    constructor(options = Utils.Const.EMPTY_OBJECT) {
        super();
        this.active = false;
        this._pointerPosition = new Vector();
        this.toViewPosition = this.toViewPosition.bind(this);
        this.toGlobalPosition = this.toGlobalPosition.bind(this);
        this._onPointerClick = this._onPointerClick.bind(this);
        this._onPointerStart = this._onPointerStart.bind(this);
        this._onPointerEnd = this._onPointerEnd.bind(this);
        if (!this.pointerConstraint) {
            this.pointerConstraint = new Constraint();
        }
        _assign(this, Scene.defaults, options);
        if (!this.objects) {
            this.objects = [];
        }
        if (!this.attachments) {
            this.attachments = [];
        }
        if (!this.effects) {
            this.effects = [];
        }
    }
    set fps(fps) {
        this.delay = 1000 / fps;
    }
    get fps() {
        return 1000 / this.delay;
    }
    set pointer(pointer) {
        const { pointerConstraint, _pointer } = this;
        if (_pointer) {
            _pointer.off('click', this._onPointerClick);
            _pointer.off('start', this._onPointerStart);
            _pointer.off('end', this._onPointerEnd);
            if (!pointer && pointerConstraint && pointerConstraint.target) {
                pointerConstraint.target.emit('dragEnd');
            }
        }
        if (this._pointer = pointer) {
            pointer.on('click', this._onPointerClick);
            pointer.on('start', this._onPointerStart);
            pointer.on('end', this._onPointerEnd);
        }
        if (pointerConstraint) {
            pointerConstraint.origin = pointer && pointer.position;
        }
    }
    get pointer() {
        return this._pointer;
    }
    _onPointerClick(position, id, event) {
        if (!this.active) {
            return;
        }
        const { pointerChecker } = this;
        if (!pointerChecker) {
            return;
        }
        const focus = this.getFocus();
        if (focus) {
            focus.emit('click', position, id, event);
        }
    }
    _draggableFilter(body) {
        return body.draggable;
    }
    _onPointerStart(position, id, event) {
        const { pointerConstraint } = this;
        if (!pointerConstraint) {
            return;
        }
        this._pointerPosition.setVector(position);
        if (pointerConstraint.target) {
            pointerConstraint.target.emit('dragEnd', position, id, event);
        }
        const focus = this.getFocus(this._draggableFilter);
        if (focus) {
            pointerConstraint.target = focus;
            pointerConstraint.targetOffset.setVector(position).minusVector(focus.position);
            focus.emit('dragStart', position, id, event);
        }
        else {
            pointerConstraint.target = _null;
        }
    }
    _onPointerEnd(position, id, event) {
        const { pointerConstraint } = this, target = pointerConstraint && pointerConstraint.target;
        if (target) {
            pointerConstraint.target = _null;
            target.emit('dragEnd', position, id, event);
            if (!target.active) {
                target.velocity.reset();
            }
        }
    }
    add(object) {
        this.objects.push(object);
        return this;
    }
    remove(object) {
        const { objects } = this, index = objects.indexOf(object);
        if (~index) {
            Utils.removeIndex(objects, index);
        }
        return this;
    }
    attach(object) {
        this.attachments.push(object);
        return this;
    }
    detach(object) {
        const { attachments } = this, index = attachments.indexOf(object);
        if (~index) {
            Utils.removeIndex(attachments, index);
        }
        return this;
    }
    use(effect) {
        this.effects.push(effect);
        return this;
    }
    disuse(effect) {
        const { effects } = this, index = effects.indexOf(effect);
        if (~index) {
            Utils.removeIndex(effects, index);
        }
        return this;
    }
    toViewPosition(position) {
        const { camera } = this;
        return camera ? camera.toViewPosition(position) : position;
    }
    toGlobalPosition(position) {
        const { camera } = this;
        return camera ? camera.toGlobalPosition(position) : position;
    }
    getFocus(filter) {
        const { pointerChecker } = this;
        if (pointerChecker) {
            const { _pointer } = this, bodies = this.objects.concat(this.attachments)
                .filter(object => object.interactive && (!filter || filter(object)));
            for (let i = bodies.length; i--;) {
                const body = bodies[i];
                if (pointerChecker(_pointer, body)) {
                    return body;
                }
            }
        }
        return _null;
    }
    drag(target) {
        const { pointer } = this;
        if (!pointer) {
            return;
        }
        const { pointerConstraint } = this;
        if (!pointerConstraint) {
            return;
        }
        const { position } = pointer;
        if (pointerConstraint.target) {
            pointerConstraint.target.emit('dragEnd');
        }
        pointerConstraint.target = target;
        if (target) {
            pointerConstraint.targetOffset.setVector(position).minusVector(target.position);
            target.emit('dragStart');
        }
    }
    update(timeScale) {
        timeScale *= this.timeScale;
        this.emit('willUpdate', timeScale);
        const deferredEffects = this.effects.filter(effect => {
            if (effect.defer) {
                return true;
            }
            else {
                effect.update(timeScale);
            }
        });
        const { pointerConstraint } = this;
        if (pointerConstraint) {
            pointerConstraint.update(timeScale);
        }
        const collidableBodies = new Array();
        this.objects.filter(object => {
            if (object.update) {
                if (object.defer) {
                    return true;
                }
                else {
                    object.update(timeScale);
                }
                if (object.category && object.collisionFilter) {
                    object.contact.clear();
                    collidableBodies.push(object);
                }
            }
        }).forEach(object => {
            object.update(timeScale);
        });
        const target = pointerConstraint && pointerConstraint.target, { _pointer } = this;
        if (target && _pointer) {
            const { position } = _pointer;
            target.velocity.setVector(position).minusVector(this._pointerPosition);
            this._pointerPosition.setVector(position);
        }
        const { collisionChecker } = this;
        if (collisionChecker) {
            Collision.check(collidableBodies, collisionChecker);
        }
        deferredEffects.forEach(effect => {
            effect.update(timeScale);
        });
        this.emit('didUpdate', timeScale);
    }
    render(renderer) {
        renderer.fill(this.background);
        const { camera } = this, { context } = renderer;
        if (camera) {
            camera.applyTo(context);
        }
        this.emit('willRender', renderer);
        this.objects.concat(this.attachments).forEach(renderable => {
            renderer.render(renderable);
        });
        this.emit('didRender', renderer);
        if (camera) {
            camera.restore(context);
        }
    }
}
Scene.defaults = {
    delay: Timer.defaults.delay,
    timeScale: 1,
    background: '#fff',
    clean: false,
    camera: _null,
    pointer: _null,
    collisionChecker: Collision.Checker.Smart,
    pointerChecker: Collision.Checker.Smart,
};

class Inspector {
    constructor(options) {
        _assign(this, Inspector.defaults, options);
        if (!this.paragraph) {
            this.paragraph = new Paragraph({
                position: Vector.of(-230, -150),
                lineHeight: 15,
                style: {
                    font: '10px Consolas',
                    textAlign: 'left',
                    textBaseline: 'top',
                    fillStyle: '#00f',
                    strokeStyle: _null,
                },
            });
        }
    }
    update(engine) {
        const { currentScene } = engine;
        this._objects = currentScene && currentScene.objects
            .concat(currentScene.attachments);
        this._camera = currentScene && currentScene.camera;
        this.paragraph.lines = this.callbacks.map(callback => callback(engine));
    }
    render(renderer) {
        const { _objects, _camera } = this, { context } = renderer;
        if (_objects) {
            if (_camera) {
                _camera.applyTo(context);
            }
            if (this.boundsStroke && _objects.length) {
                context.beginPath();
                _objects.forEach(object => {
                    const { bounds } = object;
                    if (bounds) {
                        context.rect(bounds.left, bounds.top, bounds.width, bounds.height);
                    }
                });
                context.strokeStyle = this.boundsStroke;
                context.lineWidth = this.boundsStrokeWidth;
                context.stroke();
            }
            if (this.velocityStroke) {
                const { velocityStrokeScale } = this;
                context.beginPath();
                _objects.forEach(object => {
                    const { velocity } = object;
                    if (velocity) {
                        const { position } = object, { x, y } = position;
                        context.moveTo(x, y);
                        context.lineTo(x + velocity.x * velocityStrokeScale, y + velocity.y * velocityStrokeScale);
                    }
                });
                context.strokeStyle = this.velocityStroke;
                context.lineWidth = this.velocityStrokeWidth;
                context.stroke();
            }
            if (_camera) {
                _camera.restore(context);
            }
        }
        renderer.render(this.paragraph);
    }
}
Inspector.defaults = {
    callbacks: [
        engine => {
            const { currentScene: scene } = engine;
            return `FPS: ${scene && scene.delay >= 0 ?
                (1000 / engine.timer.lastFrameDelay).toFixed(3) : 0}`;
        },
        engine => `Frame Duration: ${engine.timer.lastFrameDuration}`,
        engine => `Objects: ${engine.currentScene ? engine.currentScene.objects.length : 0}`,
        engine => `Effects: ${engine.currentScene ? engine.currentScene.effects.length : 0}`,
        engine => `Attachments: ${engine.currentScene ? engine.currentScene.attachments.length : 0}`,
        engine => `Pointer Position: ${engine.pointer.position}`,
    ],
    boundsStroke: _null,
    boundsStrokeWidth: 1,
    velocityStroke: _null,
    velocityStrokeWidth: 1,
    velocityStrokeScale: 1,
};

class Engine extends EventEmitter {
    constructor(options = Utils.Const.EMPTY_OBJECT) {
        super();
        this.inspector = _null;
        this.currentScene = _null;
        _assign(this, Engine.defaults, options);
        if (!this.timer) {
            this.timer = new Timer();
        }
        if (!this.renderer) {
            this.renderer = new Renderer();
        }
        if (!this.pointer) {
            this.pointer = new Pointer();
        }
        this.timer.on('tick', this.tick = this.tick.bind(this));
        const { renderer } = this;
        this.pointer.pretransform = renderer.toInnerPosition.bind(this.renderer);
        renderer.on('resize', () => {
            if (this.rerenderOnResize) {
                const { currentScene } = this;
                if (currentScene) {
                    const { context } = renderer;
                    this.emit('willRender', renderer);
                    context.save();
                    renderer.render(currentScene);
                    context.restore();
                    this.emit('didRender', renderer);
                }
            }
        });
    }
    createScene(options) {
        return new Scene(_assign({ pointer: this.pointer }, options));
    }
    enter(scene) {
        const { currentScene } = this;
        if (currentScene) {
            currentScene.active = false;
            currentScene.emit('exit');
        }
        const { pointer } = this;
        this.currentScene = scene;
        if (scene) {
            const { timer } = this;
            if (timer.delay !== scene.delay) {
                timer.reschedule(scene.delay);
            }
            pointer.transform = scene.toViewPosition;
            scene.active = true;
            scene.emit('enter');
            if (!timer.isRunning) {
                timer.start();
            }
        }
        else {
            pointer.transform = _null;
        }
    }
    tick(deltaTime) {
        if (deltaTime > this.maxDelay) {
            return;
        }
        const { currentScene, inspector, renderer } = this, { context } = renderer;
        if (currentScene) {
            const timeScale = deltaTime / this.baseTime;
            this.emit('willUpdate', timeScale);
            const { timer } = this;
            if (timer.delay !== currentScene.delay) {
                timer.reschedule(currentScene.delay);
            }
            currentScene.update(timeScale);
            this.emit('didUpdate', timeScale);
            this.emit('willRender', renderer);
            context.save();
            renderer.render(currentScene);
            context.restore();
            this.emit('didRender', renderer);
        }
        if (inspector) {
            inspector.update(this);
            renderer.render(inspector);
        }
    }
}
Engine.defaults = {
    baseTime: 50,
    maxDelay: 1500,
    rerenderOnResize: true,
    restoration: true,
};

export { Body, Bounds, Camera, Category, Circle, Collision, Constraint, Engine, EventEmitter, Inspector, Layer, Line, Paragraph, Particles, Pointer, Polygon, Pool, Rectangle, Renderer, Scene, Shape, Sizing, Sprite, Style, Text, Timer, Timing, Transition, Utils, Vector, Vertices };
