import { EventEmitter } from "../common/EventEmitter";
import { Vector, VectorLike } from "../geometry/Vector";
import { CategoryTag } from "./Category";
import { Bounds } from "../geometry/Bounds";
import { Renderable, RendererLike } from "../renderer/Renderer";
import { PointerEventParameters } from "../core/Pointer";
import { CollisionInfo } from "./Collision";
export interface Projection {
    min: number;
    max: number;
}
export interface BodyLike {
    isCircle: boolean;
    radius: number;
    position: Vector;
    bounds: Bounds;
    normals: ReadonlyArray<Vector>;
    getClosest(target: VectorLike): Vector;
    project(direction: Vector): Projection;
}
export declare type BodyOptions = Partial<{
    tag: CategoryTag;
    category: number;
    collisionFilter: number;
    sensorFilter: number;
    isCircle: boolean;
    active: boolean;
    interactive: boolean;
    draggable: boolean;
    defer: boolean;
    position: Vector;
    acceleration: Vector;
    velocity: Vector;
    maxSpeed: number;
    angularSpeed: number;
    maxAngularSpeed: number;
    gravity: Vector | null;
    density: number;
    mass: number;
    stiffness: number;
    slop: number;
    elasticity: number;
    friction: number;
    staticFriction: number;
    airFriction: number;
    scaleX: number;
    scaleY: number;
    rotation: number;
    fixRotation: boolean;
    radius: number;
}>;
export interface BodyEvents {
    willUpdate: number;
    didUpdate: number;
    collision: [Body, CollisionInfo];
    click: PointerEventParameters;
    dragStart: PointerEventParameters | [];
    dragEnd: PointerEventParameters | [];
}
export declare abstract class Body extends EventEmitter<BodyEvents> implements Required<BodyOptions>, BodyLike, Renderable {
    static normalPrecision: number;
    static maxStaticSpeed: number;
    static defaults: BodyOptions;
    constructor(options?: Readonly<BodyOptions>);
    readonly tag: CategoryTag;
    readonly category: number;
    readonly bounds: Bounds;
    readonly area: number;
    readonly normals: ReadonlyArray<Vector>;
    readonly scaleX: number;
    readonly scaleY: number;
    readonly rotation: number;
    readonly density: number;
    readonly mass: number;
    readonly isStatic: boolean;
    readonly speed: number;
    readonly contact: Set<Body>;
    active: boolean;
    interactive: boolean;
    draggable: boolean;
    collisionFilter: number;
    sensorFilter: number;
    isCircle: boolean;
    defer: boolean;
    position: Vector;
    acceleration: Vector;
    velocity: Vector;
    maxSpeed: number;
    angularSpeed: number;
    maxAngularSpeed: number;
    gravity: Vector | null;
    stiffness: number;
    slop: number;
    elasticity: number;
    friction: number;
    staticFriction: number;
    airFriction: number;
    fixRotation: boolean;
    radius: number;
    x: number;
    y: number;
    setDensity(density: number): this;
    setMass(mass: number): this;
    protected _setArea(area: number): void;
    activate(active?: boolean): this;
    protected _scale?(scaleX: number, scaleY: number, origin?: VectorLike): void;
    protected _rotate?(rotation: number, origin?: VectorLike): void;
    abstract getClosest(target: VectorLike): Vector;
    abstract project(direction: Vector): Projection;
    abstract render(renderer: RendererLike): void;
    setScale(scaleX: number, scaleY?: number): this;
    scale(scaleX: number, scaleY?: number, origin?: VectorLike): this;
    shrink(shrinkX: number, shrinkY?: number, origin?: VectorLike): this;
    setRotation(rotation: number, origin?: VectorLike): this;
    rotate(rotation: number, origin?: VectorLike): this;
    applyForce(force: VectorLike, point?: VectorLike): this;
    move(deltaX: number, deltaY: number): this;
    moveVector(vector: VectorLike, scale?: number): this;
    moveTo(x: number, y: number): this;
    moveToVector(vector: VectorLike): this;
    update(timeScale: number): void;
}
